import asyncio
import json
import logging

from jsonstreamer import ObjectStreamer

from .sendqueue import SendQueue
from .utils.jsonencoder import TrellioEncoder


class JSONProtocol(asyncio.Protocol):
    logger = logging.getLogger(__name__)

    def __init__(self):
        self._send_q = None
        self._connected = False
        self._transport = None
        self._obj_streamer = None
        self._pending_data = []

    @staticmethod
    def _make_frame(packet):
        string = json.dumps(packet, cls=TrellioEncoder) + ','
        return string.encode()

    def is_connected(self):
        return self._connected

    def _write_pending_data(self):
        for packet in self._pending_data:
            frame = self._make_frame(packet)
            self._transport.write(frame)
        self._pending_data.clear()

    def connection_made(self, transport):
        self._connected = True
        self._transport = transport
        try:
            self._transport.send = self._transport.write
        except:
            pass
        self._send_q = SendQueue(transport, self.is_connected)
        self.set_streamer()
        self._send_q.send()

    def set_streamer(self):
        self._obj_streamer = ObjectStreamer()
        self._obj_streamer.auto_listen(self, prefix='on_')
        self._obj_streamer.consume('[')

    def connection_lost(self, exc):
        self._connected = False
        self.logger.info('Peer closed %s', self._transport.get_extra_info('peername'))

    def send(self, packet):
        frame = self._make_frame(packet)
        self._send_q.send(frame)
        self.logger.debug('Data sent: %s', frame.decode())

    def close(self):
        self._transport.write(']'.encode())  # end the json array
        self._transport.close()

    def data_received(self, byte_data):
        string_data = byte_data.decode()
        self.logger.debug('Data received: %s', string_data)
        try:
            self._obj_streamer.consume(string_data)
        except:
            # recover from invalid data
            self.logger.exception('Invalid data received')
            self.set_streamer()

    def on_object_stream_start(self):
        raise RuntimeError('Incorrect JSON Streaming Format: expect a JSON Array to start at root, got object')

    def on_object_stream_end(self):
        del self._obj_streamer
        raise RuntimeError('Incorrect JSON Streaming Format: expect a JSON Array to end at root, got object')

    def on_array_stream_start(self):
        self.logger.debug('Array Stream started')

    def on_array_stream_end(self):
        del self._obj_streamer
        self.logger.debug('Array Stream ended')

    def on_pair(self, pair):
        self.logger.debug('Pair {}'.format(pair))
        raise RuntimeError('Received a key-value pair object - expected elements only')


class TrellioProtocol(JSONProtocol):
    def __init__(self, handler):
        super(TrellioProtocol, self).__init__()
        self._handler = handler

    def connection_made(self, transport):
        peer_name = transport.get_extra_info('peername')
        self.logger.info('Connection from %s', peer_name)
        super(TrellioProtocol, self).connection_made(transport)

    def connection_lost(self, exc):
        super(TrellioProtocol, self).connection_lost(exc)
        try:
            self._handler._handle_connection_lost()
        except:
            pass
            # self.logger.exception(str(e))

    def on_element(self, element):
        try:
            self._handler.receive(packet=element, protocol=self, transport=self._transport)
        except:
            # ignore any unhandled errors raised by handler
            self.logger.exception('api request exception')
            pass



import multiprocessing

import requests
from aiohttp.web_response import json_response

from trellio import *
from trellio.registry import Repository, Registry

processes = []


class ServiceA(TCPService):
    def __init__(self):
        super().__init__("service_a", "1", host_port=8001, host_ip='0.0.0.0')

    @api
    async def echo(self, data):
        return data


class ServiceClientA(TCPServiceClient):
    def __init__(self):
        super(ServiceClientA, self).__init__("service_a", "1")

    @request
    def echo(self, data):
        return locals()


class ServiceB(HTTPService):
    def __init__(self):
        super().__init__("service_b", "1", host_port=8000, host_ip='0.0.0.0')
        self._client_a = ServiceClientA()

    @get(path="/{data}/")
    async def get_echo(self, request):
        data = request.match_info.get('data')
        d = await self._client_a.echo(data)
        return json_response(text=d)


class ServiceBTCP(TCPService):
    def __init__(self):
        super().__init__("service_b", "1", host_port=8003, host_ip='0.0.0.0')
        self._client_a = ServiceClientA()

    @api
    async def get_echo(self, data):
        data = request.match_info.get('data')
        d = await self._client_a.echo(data)
        return d


def start_registry():
    repository = Repository()
    registry = Registry('0.0.0.0', 4500, repository)
    registry.start()


def start_service_a():
    Host.configure(http_port=8002, tcp_port=8001, tcp_host='0.0.0.0', service_name="service_a", service_version="1")
    Host.attach_tcp_service(ServiceA())
    Host.run()


def start_service_b():
    client_a = ServiceClientA()
    service_b = ServiceB()
    tcp_service = ServiceBTCP()

    Host.configure(http_port=8000, tcp_port=8003, http_host='0.0.0.0', service_name="service_b", service_version="1",
                   registry_host='0.0.0.0', registry_port=4500)

    service_b.clients = [client_a]
    tcp_service.clients = [client_a]
    Host.attach_http_service(service_b)
    Host.attach_tcp_service(tcp_service)
    Host.run()


def setup_module():
    global processes
    for target in [start_registry, start_service_a, start_service_b]:
        p = multiprocessing.Process(target=target)
        p.start()
        processes.append(p)

        # allow the subsystems to start up.
        # sleep for awhile
        import time
        time.sleep(1)


def restart_service_a():
    processes[1].terminate()
    p = multiprocessing.Process(target=start_service_a)
    p.start()
    processes[1] = p
    import time
    time.sleep(1)


def teardown_module():
    for p in processes:
        p.terminate()


def test_service_b():
    url = 'http://0.0.0.0:8000/blah/'
    r = requests.get(url)
    assert r.text == 'blah'
    assert r.status_code == 200


if __name__ == "__main__":
    setup_module()
    restart_service_a()
    test_service_b()



__all__ = ['HTTPView', 'TCPView']

from again.utils import unique_hex

from .utils.helpers import default_preflight_response
from .utils.ordered_class_member import OrderedClassMembers


class BaseView:
    '''base class for views'''
    _host = None

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, host):
        self._host = host


class BaseHTTPView(BaseView, metaclass=OrderedClassMembers):
    '''base class for HTTP views'''
    middlewares = []

    def __init__(self):
        super(BaseHTTPView, self).__init__()


class HTTPView(BaseHTTPView):
    def __init__(self, allow_cross_domain=True,
                 preflight_response=default_preflight_response):
        super(HTTPView, self).__init__()
        self._allow_cross_domain = allow_cross_domain
        self._preflight_response = preflight_response

    @property
    def cross_domain_allowed(self):
        return self._allow_cross_domain

    @property
    def preflight_response(self):
        return self._preflight_response


class BaseTCPView(BaseView):
    '''base class for TCP views'''

    def __init__(self):
        super(BaseTCPView, self).__init__()


class TCPView(BaseTCPView):
    def __init__(self):
        super(TCPView, self).__init__()

    @staticmethod
    def _make_response_packet(request_id: str, from_id: str, entity: str, result: object, error: object,
                              failed: bool, old_api=None, replacement_api=None):
        from .services import _Service
        if failed:
            payload = {'request_id': request_id, 'error': error, 'failed': failed}
        else:
            payload = {'request_id': request_id, 'result': result}
        if old_api:
            payload['old_api'] = old_api
            if replacement_api:
                payload['replacement_api'] = replacement_api
        packet = {'pid': unique_hex(),
                  'to': from_id,
                  'entity': entity,
                  'type': _Service._RES_PKT_STR,
                  'payload': payload}
        return packet



import asyncio
import datetime
import logging
import logging.config
import sys
from functools import partial, wraps
from logging import Handler
from queue import Queue
from threading import Thread

import yaml
from pythonjsonlogger import jsonlogger

RED = '\033[91m'
BLUE = '\033[94m'
BOLD = '\033[1m'
END = '\033[0m'


class CustomTimeLoggingFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):  # noqa
        """
        Overrides formatTime method to use datetime module instead of time module
        to display time in microseconds. Time module by default does not resolve
        time to microseconds.
        """
        if datefmt:
            s = datetime.datetime.now().strftime(datefmt)
        else:
            t = datetime.datetime.now().strftime(self.default_time_format)
            s = self.default_msec_format % (t, record.msecs)
        return s


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def __init__(self, *args, **kwargs):
        self.extrad = kwargs.pop('extrad', {})
        super().__init__(*args, **kwargs)

    def add_fields(self, log_record, record, message_dict):
        message_dict.update(self.extrad)
        super().add_fields(log_record, record, message_dict)


def patch_async_emit(handler: Handler):
    base_emit = handler.emit
    queue = Queue()

    def loop():
        while True:
            record = queue.get()
            try:
                base_emit(record)
            except:
                print(sys.exc_info())

    def async_emit(record):
        queue.put(record)

    thread = Thread(target=loop)
    thread.daemon = True
    thread.start()
    handler.emit = async_emit
    return handler


def patch_add_handler(logger):
    base_add_handler = logger.addHandler

    def async_add_handler(handler):
        async_handler = patch_async_emit(handler)
        base_add_handler(async_handler)

    return async_add_handler


DEFAULT_CONFIG_YAML = """
    # logging config

    version: 1
    disable_existing_loggers: False
    handlers:
        stream:
            class: logging.StreamHandler
            level: INFO
            formatter: ctf
            stream: ext://sys.stdout

        stats:
            class: logging.StreamHandler
            level: INFO
            formatter: cjf
            stream: ext://sys.stdout

        service:
            class: logging.StreamHandler
            level: INFO
            formatter: ctf
            stream: ext://sys.stdout

    formatters:
        ctf:
            (): trellio.utils.log.CustomTimeLoggingFormatter
            format: '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            datefmt: '%Y-%m-%d %H:%M:%S,%f'

        cjf:
            (): trellio.utils.log.CustomJsonFormatter
            format: '{ "timestamp":"%(asctime)s", "message":"%(message)s"}'
            datefmt: '%Y-%m-%d %H:%M:%S,%f'

    root:
        handlers: [stream]
        level: INFO

    loggers:
        registry:
            handlers: [stream]
            level: INFO

        stats:
            handlers: [stats]
            level: INFO

    """


def setup_logging(_):
    try:
        with open('config_log.json', 'r') as f:
            config_dict = yaml.load(f.read())
    except:
        config_dict = yaml.load(DEFAULT_CONFIG_YAML)

    logging.getLogger('asyncio').setLevel(logging.DEBUG)
    logger = logging.getLogger()
    logger.propagate = False
    logger.handlers = []
    logger.addHandler = patch_add_handler(logger)
    logging.config.dictConfig(config_dict)


def log(fn=None, logger=logging.getLogger(), debug_level=logging.DEBUG):
    """
    logs parameters and result - takes no arguments
    """
    if fn is None:
        return partial(log, logger=logger, debug_level=debug_level)

    @wraps(fn)
    def func(*args, **kwargs):
        arg_string = ""
        for i in range(0, len(args)):
            var_name = fn.__code__.co_varnames[i]
            if var_name not in ['self', 'cls']:
                arg_string += var_name + ":" + str(args[i]) + ","
        arg_string = arg_string[0:len(arg_string) - 1]
        string = (RED + BOLD + '>> ' + END + 'Calling {0}({1})'.format(fn.__name__, arg_string))
        if len(kwargs):
            string = (
                RED + BOLD + '>> ' + END + 'Calling {0} with args {1} and kwargs {2}'.format(fn.__name__, arg_string,
                                                                                             kwargs))
        logger.log(debug_level, string)
        wrapped_fn = fn
        if not asyncio.iscoroutine(fn):
            wrapped_fn = asyncio.coroutine(fn)
        try:
            result = yield from wrapped_fn(*args, **kwargs)
            string = BLUE + BOLD + '<< ' + END + 'Return {0} with result :{1}'.format(fn.__name__, result)
            logger.log(debug_level, string)
            return result
        except Exception as e:
            string = (RED + BOLD + '>> ' + END + '{0} raised exception :{1}'.format(fn.__name__, str(e)))
            logger.log(debug_level, string)
            raise e

    return func


def logx(supress_args=[], supress_all_args=False, supress_result=False, logger=logging.getLogger(),
         debug_level=logging.DEBUG):
    """
    logs parameters and result
    takes arguments
        supress_args - list of parameter names to supress
        supress_all_args - boolean to supress all arguments
        supress_result - boolean to supress result
        receiver - custom logging function which takes a string as input; defaults to logging on stdout
    """

    def decorator(fn):
        def func(*args, **kwargs):
            if not supress_all_args:
                arg_string = ""
                for i in range(0, len(args)):
                    var_name = fn.__code__.co_varnames[i]
                    if var_name != "self" and var_name not in supress_args:
                        arg_string += var_name + ":" + str(args[i]) + ","
                arg_string = arg_string[0:len(arg_string) - 1]
                string = (RED + BOLD + '>> ' + END + 'Calling {0}({1})'.format(fn.__name__, arg_string))
                if len(kwargs):
                    string = (
                        RED + BOLD + '>> ' + END + 'Calling {0} with args {1} and kwargs {2}'.format(
                            fn.__name__,
                            arg_string, kwargs))
                logger.log(debug_level, string)

            wrapped_fn = fn
            if not asyncio.iscoroutine(fn):
                wrapped_fn = asyncio.coroutine(fn)
            result = yield from wrapped_fn(*args, **kwargs)

            if not supress_result:
                string = BLUE + BOLD + '<< ' + END + 'Return {0} with result : {1}'.format(fn.__name__, result)
                logger.log(debug_level, string)
            return result

        return func

    return decorator



import warnings
from functools import wraps


def deprecated(func):
    """
    Generates a deprecation warning
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        msg = "'{}' is deprecated".format(func.__name__)
        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    return wrapper



class InvalidSignalType(Exception):
    pass

class BaseSignal:
    _registry_list = []

    @classmethod
    def register(cls, to_register, soft=False):
        cls._registry_list.append([to_register, soft])

    @classmethod
    async def _run(cls, *args, **kwargs):
        print('service ready run called')
        for i in cls._registry_list:
            try:
                await i[0](*args, **kwargs)
            except Exception as e:
                if not i[1]:
                    raise e


class ServiceReady(BaseSignal):
    _registry_list = []


def register(signal_type,soft=False):

    def decorator(receiver):
        if not issubclass(signal_type, BaseSignal):
            raise InvalidSignalType
        signal_type.register(receiver,soft)
        return receiver
    return decorator




from trellio import TCPService, api, Host


class ServiceTest(TCPService):
    def __init__(self):
        super().__init__('test_service', '1')

    @api
    async def echo(self):
        return "echo"


if __name__ == "__main__":
    test_service = ServiceTest()
    Host.configure(tcp_port=8001)

    Host.attach_tcp_service(test_service)
    Host.run()
